/*
 * Licensed to Elasticsearch under one or more contributor
 * license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright
 * ownership. Elasticsearch licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package co.elastic.blog.elasticsearch.example.realm;

import org.elasticsearch.common.logging.Loggers;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.Signature;
import java.security.SignatureException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;

/**
 * Responsible for verifying a cryptographic signature against a configured public key.
 * The key is provided during construction as a {@link Certificate}, and signatures are
 * checked with the {@link #verify(String, byte[])} method.
 * Signatures must use the "SHA256withRSA" algorithm.
 */
public class SignatureVerifier {

    private final Certificate certificate;

    public SignatureVerifier(Path certificatePath) throws IOException, CertificateException {
        this.certificate = readCertificate(certificatePath);
        Loggers.getLogger(getClass()).info("Using [{}] [{}] certificate from path [{}]",
                certificate.getType(), certificate.getPublicKey().getAlgorithm(), certificatePath);
    }

    /**
     * Reads a X.509 Certificate from a provided Path.
     * This relies on the JRE's builtin support for reading certificates, so it can handle various encoding
     * formats including PEM and raw DER.
     */
    private Certificate readCertificate(Path path) throws IOException, CertificateException {
        try (InputStream input = Files.newInputStream(path)) {
            final CertificateFactory factory = CertificateFactory.getInstance("X.509");
            return factory.generateCertificate(input);
        }
    }

    /**
     * Verifies the {@code signature} is valid for the provided {@code input} and the pre-configured {@code certificate}.
     */
    public void verify(String input, byte[] signature) throws InvalidKeyException, NoSuchAlgorithmException, SignatureException {
            final Signature sig = Signature.getInstance("SHA256withRSA");
        sig.initVerify(certificate);
        sig.update(input.getBytes(StandardCharsets.UTF_8));
        if (sig.verify(signature) == false) {
            throw new SignatureException("Failed to verify signature");
        }
    }
}
