/*
 * Licensed to Elasticsearch under one or more contributor
 * license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright
 * ownership. Elasticsearch licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package co.elastic.blog.elasticsearch.example.realm;

import com.google.gson.Gson;

import java.security.AccessController;
import java.security.PrivilegedAction;

/**
 * Represents information about a user in the webstore.
 * Note: A claim cannot be assumed to be true and accurate unless it is accompanied by a corresponding signature.
 */
public class WebStoreClaim {

    private String principal;
    private Long storeId;
    private String role;

    /**
     * Called by gson in {@link #parseFromJson(String)}
     */
    @SuppressWarnings("unused")
    public WebStoreClaim() {
    }

    /**
     * @return principal (username) for the claim
     */
    public String getPrincipal() {
        return principal;
    }

    /**
     * @return The identifier of the store to which this user has access
     */
    public Long getStoreId() {
        return storeId;
    }

    /**
     * @return The role that this user has on the given store. Should be one of "reader" or "writer"
     */
    public String getRole() {
        return role;
    }

    /**
     * Called by gson in {@link #parseFromJson(String)}
     */
    @SuppressWarnings("unused")
    public void setPrincipal(String principal) {
        this.principal = principal;
    }

    /**
     * Called by gson in {@link #parseFromJson(String)}
     */
    @SuppressWarnings("unused")
    public void setStoreId(Long storeId) {
        this.storeId = storeId;
    }

    /**
     * Called by gson in {@link #parseFromJson(String)}
     */
    @SuppressWarnings("unused")
    public void setRole(String role) {
        this.role = role;
    }

    /**
     * Parse a JSON string into a claim object.
     */
    static WebStoreClaim parseFromJson(String claimJson) {
        // Gson uses reflection in order to find and set the bean properties on WebStoreClaim.
        // The plugin security manager prevents this access by default so we must add the required permission
        // to "plugin-security.policy" and then wrap this call in a "doPrivileged" block
        return AccessController.doPrivileged((PrivilegedAction<WebStoreClaim>)
                // Elasticsearch ships with various classes and methods for parsing JSON (see the XContent interface)
                // This example uses Gson in order to demonstrate the use of third-party dependencies in a security extension
                () -> new Gson().fromJson(claimJson, WebStoreClaim.class)
        );
    }
}
