/*
 * Licensed to Elasticsearch under one or more contributor
 * license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright
 * ownership. Elasticsearch licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package co.elastic.blog.elasticsearch.example.realm;

import org.apache.logging.log4j.message.ParameterizedMessage;
import org.elasticsearch.action.ActionListener;
import org.elasticsearch.common.Strings;
import org.elasticsearch.common.settings.Setting;
import org.elasticsearch.common.util.concurrent.ThreadContext;
import org.elasticsearch.common.util.set.Sets;
import org.elasticsearch.watcher.FileChangesListener;
import org.elasticsearch.watcher.FileWatcher;
import org.elasticsearch.watcher.ResourceWatcherService;
import org.elasticsearch.xpack.core.security.authc.AuthenticationResult;
import org.elasticsearch.xpack.core.security.authc.AuthenticationToken;
import org.elasticsearch.xpack.core.security.authc.Realm;
import org.elasticsearch.xpack.core.security.authc.RealmConfig;
import org.elasticsearch.xpack.core.security.authc.RealmSettings;
import org.elasticsearch.xpack.core.security.user.User;

import java.io.IOException;
import java.nio.file.Path;
import java.security.GeneralSecurityException;
import java.security.cert.CertificateException;
import java.util.Base64;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * A Customer X-Pack Security Realm for authenticating users using our hypothetical eCommerce platform protocol.
 */
public class WebStoreRealm extends Realm {

    /**
     * The "type" of our realm. This corresponds to the realm's @{code type} in elasticsearch.yml.
     */
    public static final String TYPE = "webstore";

    /**
     * The name of header that contains the "claims" (see {@link WebStoreClaim}) in JSON format.
     */
    public static final String CLAIMS_HEADER = "x-web-store-claims";

    /**
     * The name of header that contains the "signature" (see {@link SignatureVerifier}) in Base64 encoding.
     */
    public static final String SIGNATURE_HEADER = "x-web-store-sig";

    /**
     * The setting for configuring the "certificate" in elasticsearch.yml.
     * This is registered as being a string, and then is resolved as a Path when it is read.
     * Realm settings are always {@link Setting.Property#NodeScope}, as realms are a per-node configuration, not a per-cluster one.
     */
    private static final Setting<String> CERTIFICATE_PATH_SETTING = Setting.simpleString("certificate", Setting.Property.NodeScope);

    /**
     * The handler for verifying the claims signature. This is volatile as it is automatically recreated whenever the certificate file is
     * modified.
     */
    private volatile SignatureVerifier signatureVerifier;

    public WebStoreRealm(RealmConfig config, ResourceWatcherService resourceWatcherService) throws IOException, CertificateException {
        super(TYPE, config);
        final Path certificatePath = getCertificatePath(config);
        this.signatureVerifier = new SignatureVerifier(certificatePath);
        watchCertificateFileForChanges(certificatePath, resourceWatcherService);
    }

    /**
     * Find the {@link Path} to a certificate file based on the provided configuration
     */
    private Path getCertificatePath(RealmConfig config) {
        final String certSetting = CERTIFICATE_PATH_SETTING.get(config.settings());
        if (Strings.isNullOrEmpty(certSetting)) {
            throw new IllegalStateException(
                    "The [" + RealmSettings.getFullSettingKey(config, CERTIFICATE_PATH_SETTING) + "] setting is required");
        }
        // Look for configuration files in the same directory as the core `elasticsearch.yml` config file
        return config.env().configFile().resolve(certSetting);
    }

    /**
     * Uses the {@link ResourceWatcherService} to track changes to the configured certificate file.
     * If the file changes, then the certificate is reloaded.
     *
     * @see #reloadCertificate(Path)
     */
    private void watchCertificateFileForChanges(Path certificatePath, ResourceWatcherService resourceWatcherService) throws IOException {
        final FileWatcher fileWatcher = new FileWatcher(certificatePath);
        resourceWatcherService.add(fileWatcher, ResourceWatcherService.Frequency.HIGH);
        fileWatcher.addListener(new FileChangesListener() {
            @Override
            public void onFileCreated(Path file) {
                reloadCertificate(certificatePath);
            }

            @Override
            public void onFileChanged(Path file) {
                reloadCertificate(certificatePath);
            }
        });
    }

    /**
     * Reloads the certificate from the provider path by constructing a new {@link SignatureVerifier}
     */
    private void reloadCertificate(Path certificatePath) {
        try {
            this.signatureVerifier = new SignatureVerifier(certificatePath);
            logger.info("Reloaded certificate {}", certificatePath);
        } catch (IOException | CertificateException e) {
            logger.warn(new ParameterizedMessage("Cannot read certificate {}, continuing to use old certificate", certificatePath), e);
        }
    }

    /**
     * Generate a token object for the current thread, if-and-only-if the provided {@link ThreadContext} contains credentials
     * in a form that this realm can handle.
     *
     * @see #supports(AuthenticationToken)
     */
    @Override
    public WebStoreToken token(ThreadContext threadContext) {
        final String json = threadContext.getHeader(CLAIMS_HEADER);
        final String sig = threadContext.getHeader(SIGNATURE_HEADER);
        if (Strings.hasText(json) && Strings.hasText(sig)) {
            // If both headers exist (with content), then extract their values into a token object
            return new WebStoreToken(json, signatureBytes(sig));
        } else {
            return null;
        }
    }

    /**
     * @return {@code true} if the provided token object is in a form that this realm can handle.
     * @see #token(ThreadContext)
     */
    @Override
    public boolean supports(AuthenticationToken authenticationToken) {
        // This realm only supports our custom token type.
        return authenticationToken instanceof WebStoreToken;
    }

    /**
     * The signature header is Base64 encoded, so we decode it to a byte[]
     */
    private byte[] signatureBytes(String sig) {
        return Base64.getDecoder().decode(sig);
    }

    /**
     * Attempt to authenticate the user identified by the provided token, and pass an appropriate result through the
     * {@link ActionListener}.
     */
    @Override
    public void authenticate(AuthenticationToken authenticationToken, ActionListener<AuthenticationResult> actionListener) {
        // This should aways be true due to our implementation of supports()
        // We should never get passed a token that we don't support.
        assert authenticationToken instanceof WebStoreToken;
        try {
            // Get a claim object from the token, ensuring that the signature has been verified
            final WebStoreClaim claim = ((WebStoreToken) authenticationToken).getClaim(signatureVerifier);
            // Construct a User object, and return it as a successful result
            actionListener.onResponse(AuthenticationResult.success(buildUser(claim)));
        } catch (GeneralSecurityException e) {
            // This indicates a signature problem - the signature for the claims is not correct.
            // However, there could be another webstore realm configured with a different signing certificate, which
            // will validate successfully, so we want to return an "unsuccessful" result rather than a fatal failure.
            // This "unsucessful" result indicates that _this_ realm could not authenticate the user, but another realm
            // might be able to. We provide an error message so that if _no_ realms can authenticate successfully the
            // core AuthenticationService will provide appropriate diagnostic logging for the failure.
            actionListener.onResponse(AuthenticationResult.unsuccessful("Cannot verify claim", e));
        } catch (Exception e) {
            // Some other exception that we didn't expect.
            // We treat this as a critical failure.
            actionListener.onFailure(e);
        }
    }

    /**
     * Build a user with suitable roles and metadata based on a provided {@link WebStoreClaim}
     */
    private User buildUser(WebStoreClaim claim) {
        final String[] roles = {
                "webstore-" + claim.getStoreId() + "-" + claim.getRole(),
                "webstore"
        };
        final Map<String, Object> metadata = new HashMap<>();
        metadata.put("webstore-store", claim.getStoreId());
        metadata.put("webstore-role", claim.getRole());
        return new User(claim.getPrincipal(), roles, null, null, metadata, true);
    }

    @Override
    public void lookupUser(String userid, ActionListener<User> actionListener) {
        // This method will be called to find an existing user by "userid" in order to support the "run_as" functionality of
        // x-pack, but we just respond with "null" because this realm cannot lookup a user by id - we need to read the right headers
        actionListener.onResponse(null);

    }

    public static Set<Setting<?>> getSettings() {
        // The settings that are allowed for our realm. Called by our security extension
        return Sets.newHashSet(CERTIFICATE_PATH_SETTING);
    }
}
