/*
 * Licensed to Elasticsearch under one or more contributor
 * license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright
 * ownership. Elasticsearch licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package co.elastic.blog.elasticsearch.example.roles;

import org.elasticsearch.action.ActionListener;
import org.elasticsearch.common.settings.Settings;
import org.elasticsearch.watcher.ResourceWatcherService;
import org.elasticsearch.xpack.core.security.authz.RoleDescriptor;

import java.util.HashSet;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A {@link org.elasticsearch.xpack.core.security.SecurityExtension#getRolesProviders(Settings, ResourceWatcherService) roles provider}
 * that dynamically constructs the required {@link RoleDescriptor} for the store specific roles in our eCommerce platform.
 */
public class WebStoreRolesProvider {

    // Role names are of the form "webstore-{store-id}-{role} where role must be one of "reader or "writer"
    private static final Pattern ROLE_PATTERN = Pattern.compile("webstore-(\\d+)-(reader|writer)");

    /**
     * For any role name in {@code names}, that matches {@link #ROLE_PATTERN}, constructs a
     * {@link RoleDescriptor} representing that role, and returns all such roles through the
     * {@code listener}.
     *
     * The signature for this method matches {@link java.util.function.BiConsumer}, so that it can be reference by
     * {@link org.elasticsearch.xpack.core.security.SecurityExtension#getRolesProviders(Settings, ResourceWatcherService)}.
     */
    public void lookup(Set<String> names, ActionListener<Set<RoleDescriptor>> listener) {
        try {
            Set<RoleDescriptor> roles = new HashSet<>();
            for (String name : names) {
                final Matcher matcher = ROLE_PATTERN.matcher(name);
                if (matcher.matches()) {
                    roles.add(buildRole(name, Long.parseLong(matcher.group(1)), matcher.group(2)));
                }
            }
            listener.onResponse(roles);
        } catch (Exception e) {
            listener.onFailure(e);
        }
    }

    /**
     * Builds a {@link RoleDescriptor} for the provided arguments
     * @param name The name of the role to construct
     * @param storeId The identifier of the store to which this role has access
     * @param access The access level - one of "reader" or "writer"
     */
    private RoleDescriptor buildRole(String name, long storeId, String access) {
        final RoleDescriptor.IndicesPrivileges.Builder indexBuilder = RoleDescriptor.IndicesPrivileges.builder();
        indexBuilder.indices("webstore-" + storeId);
        if (access.equals("reader")) {
            indexBuilder.privileges("read");
        } else if (access.equals("writer")) {
            indexBuilder.privileges("write");
        } else {
            throw new IllegalArgumentException("Unsupported access type " + access);
        }
        return new RoleDescriptor(name, new String[]{"monitor"}, new RoleDescriptor.IndicesPrivileges[]{indexBuilder.build()}, null);
    }
}
